import socket
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape

from .settings import CONF_DIR, EXAMPLE_PURE_HTML, PKG_DIR, PLATFORM

try:
    import pyqrcode

    PYQRCODE_IMPORT = True
except ImportError:
    PYQRCODE_IMPORT = False


def generate_qr(
    port: int = 8888,
    username: str = "thqm",
    password: str = None,
    qr_path: Path = PKG_DIR / "styles/default/static/qr_code.svg",
):
    """Generate the qrcode containing login credential if provided.
    Requires 'pyqrcode'.

    Args:
        port: port number.
        username: login username.
        password: login password.
        qr_path: path where to store the generated qr code svg.

    Returns:
        A pyqrcode.qr object containing the url and login credentials if
        provided.
    """

    if password is not None:
        qr_url = f"http://{username}:{password}@{get_ip()}:{port}/"
    else:
        qr_url = f"http://{get_ip()}:{port}/"
    qr = pyqrcode.create(qr_url)
    qr.svg(qr_path, module_color="#000000", background="#ffffff")
    return qr


def get_ip() -> str:
    """Gets host local ip.

    Returns:
        LAN ip.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))  # checks if device is connected to internet
    ip = s.getsockname()[0]
    s.close()
    return ip


def echo(msg: str):
    """Print and flush.
    """
    print(msg, flush=True)


def check_base_dir(custom_folder: Path) -> bool:
    """Check to make sure the minimum requirements are met.

    Return:
        True if templates/index.html exists in custom_folder, False otherwise.
    """
    return (
        (custom_folder / "static").is_dir()
        and (custom_folder / "templates").is_dir()
        and (custom_folder / "templates/index.html").is_file()
    )


def get_styles() -> list:
    """Get the available styles.
    """
    return [
        style
        for style in list((PKG_DIR / "styles").glob("*")) + list(CONF_DIR.glob("*"))
        if style.is_dir() and check_base_dir(style)
    ]


def style_base_dir(style: str) -> Path:
    """Get style path from name.
    """
    styles = {style_path.name: style_path for style_path in get_styles()}
    return styles[style]


def create_jinja_env(custom_folder: Path = None) -> Environment:
    """Create a jinja environment for the custom path. custom_folder should contain a
    templates/index.html file.
    """
    if custom_folder is not None:
        if not check_base_dir(custom_folder):
            raise FileNotFoundError(
                f"'{custom_folder}' not valid, must contain 'templates/index.html'."
            )
        else:
            env = Environment(
                loader=FileSystemLoader(custom_folder / "templates"),
                autoescape=select_autoescape(["html", "xml"]),
            )
    else:
        env = Environment(
            loader=PackageLoader("thqm", "templates"),
            autoescape=select_autoescape(["html", "xml"]),
        )
    return env


def init_conf_folder():
    folders = [
        CONF_DIR,
        CONF_DIR / "pure_html",
        CONF_DIR / "pure_html/static",
        CONF_DIR / "pure_html/templates",
    ]
    if not CONF_DIR.is_dir():
        for folder in folders:
            if not folder.is_dir():
                folder.mkdir()
        (CONF_DIR / "pure_html/templates/index.html").write_text(EXAMPLE_PURE_HTML)



import shutil
import sys
import threading
from base64 import b64encode
from copy import copy
from http.server import BaseHTTPRequestHandler, HTTPServer
from io import BytesIO
from pathlib import Path
from typing import Text
from urllib.parse import unquote

from .settings import PKG_DIR
from .utils import PYQRCODE_IMPORT, create_jinja_env, echo, style_base_dir


def handler_factory(
    jinja_template_rendered: BytesIO,
    base_dir: Path,
    events: list = [],
    username: str = "thqm",
    password: str = None,
    oneshot: bool = False,
):
    """Create a HTTPHandler class with the desired properties.

    Args:
        jinja_template_rendered: BytesIO object of the rendered template.
        base_dir: directory containing the static/ and templates/ folders.
        events: allowed events.
        username: basic auth username.
        password: basic auth password.
        oneshot: stop server after first click.

    Returns:
        HTTPHandler class.
    """

    class HTTPHandler(BaseHTTPRequestHandler):

        extensions_map = {
            ".html": "text/html",
            "": "application/octet-stream",  # Default
            ".css": "text/css",
            ".js": "text/javascript",
            ".png": "image/png",
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".svg": "image/svg+xml",
        }

        def __init__(self, *args, **kwargs):
            self.events = events
            self.require_login = password is not None
            self._auth = b64encode(f"{username}:{password}".encode()).decode()
            super().__init__(*args, **kwargs)

        def _do_GET(self):
            f = self.send_head()
            if f:
                self.copyfile(f, self.wfile)
                f.close()

        def do_GET(self):
            """Serve a GET request.
            """
            if self.require_login:
                if self.headers.get("Authorization") == "Basic " + self._auth:
                    self._do_GET()
                else:
                    self.do_HEADAUTH()
            else:
                self._do_GET()

        def do_HEAD(self):
            """Serve a HEAD request.
            """
            f = self.send_head()
            if f:
                f.close()

        def do_HEADAUTH(self):
            """Handle the authentication in the header.
            """
            self.send_response(401)
            self.send_header("WWW-Authenticate", 'Basic realm="thqm"')
            self.send_header("Content-type", "text/html")
            self.end_headers()

        def send_head(self):
            """Common code for GET and HEAD commands.

            This sends the response code and MIME headers.

            Return value is either a file object (which has to be copied
            to the outputfile by the caller unless the command was HEAD,
            and must be closed by the caller under all circumstances), or
            None, in which case the caller has nothing further to do.
            """
            path = self.translate_path(self.path)
            f = None
            ctype = None

            if self.get_query(self.path) == "shutdown":
                # shutdown server
                self.shutdown()
                return
            elif path in (base_dir / e for e in self.events):
                # If event
                echo(path.relative_to(base_dir))
                if oneshot:
                    # shutdown after print
                    self.shutdown()
                    return
                self.send_response(302)
                self.send_header("Location", "/")
                self.end_headers()
                return
            elif path == base_dir:
                # if control panel
                f = copy(jinja_template_rendered)
                ctype = "text/html"
            elif (base_dir / "static") in path.parents:
                # if anything else
                try:
                    f = open(path, "rb")
                except IOError:
                    return
            else:
                return
            if not ctype:
                ctype = self.guess_type(path)
            self.send_response(200)
            self.send_header("Content-type", ctype)
            self.end_headers()
            return f

        def translate_path(self, path: str) -> Path:
            """Translate a /-separated PATH to the local filename syntax.
            """
            # abandon query parameters
            path = path.split("?", 1)[0]
            path = path.split("#", 1)[0]
            # remove first /
            return base_dir / unquote(path)[1:]

        def get_query(self, path: str) -> Path:
            """Get the first query parameter.
            """
            path = path.split("?", 1)
            if len(path) > 1:
                return path[1]

        def shutdown(self):
            """Shutdown the server.
            """
            killer = threading.Thread(target=self.server.shutdown)
            killer.start()

        def copyfile(self, source, outputfile):
            """Copy all data between two file objects.

            The SOURCE argument is a file object open for reading
            (or anything with a read() method) and the DESTINATION
            argument is a file object open for writing (or
            anything with a write() method).

            The only reason for overriding this would be to change
            the block size or perhaps to replace newlines by CRLF
            -- note however that this the default server uses this
            to copy binary data as well.
            """
            shutil.copyfileobj(source, outputfile)

        def guess_type(self, path: Path) -> str:
            """Guess the type of a file.

            Argument is a PATH (a filename).

            Return value is a string of the form type/subtype,
            usable for a MIME Content-type header.

            The default implementation looks the file's extension
            up in the table self.extensions_map, using application/octet-stream
            as a default; however it would be permissible (if
            slow) to look inside the data to make a better guess.
            """
            ext = path.suffix.lower()
            return self.extensions_map.get(ext, self.extensions_map[""])

        def log_message(self, *args, **kwargs):
            """Disable all prints.
            """
            pass

    return HTTPHandler


def start_server(
    events: list = [],
    port: int = 8901,
    username: str = "thqm",
    password: str = None,
    oneshot: bool = False,
    base_dir: Path = PKG_DIR / "styles/default",
    **jinja_template_kwargs,
):
    """Renders the template and starts the server.

    Args:
        events: list of events.
        port: port number on which to run the server.
        username: login username.
        password: login password.
        oneshot: stop server after first click.
        base_dir: server base directory, must have a templates/index.html file.
        **jinja_template_kwargs: template args.
    """

    jinja_env = create_jinja_env(base_dir)
    template = jinja_env.get_template("index.html")

    handler = handler_factory(
        base_dir=base_dir,
        events=events,
        username=username,
        password=password,
        oneshot=oneshot,
        jinja_template_rendered=BytesIO(
            template.render(events=events, **jinja_template_kwargs).encode("utf8")
        ),
    )

    server_address = ("", port)
    httpd = HTTPServer(server_address, handler)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        httpd.shutdown()
        raise



import os
import sys
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape


def get_conf_path() -> Path:
    """Platform agnostic config directory.

    Returns:
        config folder path.
    """
    if PLATFORM.startswith("win"):
        key = "LOCALAPPDATA"
        folder = os.environ.get(key)
        if folder is None:
            folder = Path.home()
        return folder / "thqm"
    if PLATFORM == "darwin":
        return Path("~/Library/Application Support").expanduser() / "thqm"
    return (
        Path(os.environ.get("XDG_CONFIG_HOME", Path("~/.config").expanduser())) / "thqm"
    )


PLATFORM = sys.platform
PKG_DIR = Path(__file__).absolute().parent
CONF_DIR = get_conf_path()
EXAMPLE_PURE_HTML = """\
{# thqm template
    Pure html, no bells or whistles.

    args:
        title (str): page title.
        qrcode_button (bool): show qrcode button.
        shutdown_button (bool): show shutdown button.
        events (list): list of strings.
#}
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <!-- <meta name="viewport" content="width=device-width, initial-scale=1" /> -->
    <title>{{ title }}</title>
  </head>

  <body>
    <header>
      <div>
        <h2>{{ title }}</h1>
        {% if qrcode_button %}
          <a href='static/qr_code.svg'>qrcode</a>
        {% endif %}
        {% if shutdown_button %}
          <a href="./?shutdown")>shutdown</a>
        {% endif %}
      </div>
    </header>
    <hr/>
    {% for e in events %}
      <a href='./{{ e }}'>
        <pre><code>{{ e|safe }}</code></pre>
      </a>
    {% endfor %}
  </body>
</html>"""



#!/usr/bin/env python
import argparse
import sys

from .server import start_server
from .utils import (
    PYQRCODE_IMPORT,
    echo,
    generate_qr,
    get_styles,
    init_conf_folder,
    style_base_dir,
)


def main():
    """thqm cli.
    """
    parser = argparse.ArgumentParser(
        prog="thqm",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Remote command execution made easy.",
    )
    parser.add_argument("-p", "--port", type=int, default=8901, help="Port number.")
    parser.add_argument(
        "-q",
        "--show-qrcode",
        action="store_true",
        default=False,
        help='Show the qrcode in terminal, requires "pyqrcode".',
    )
    parser.add_argument(
        "-pw", "--password", default=None, type=str, help="Authentication password."
    )
    parser.add_argument(
        "-u",
        "--username",
        default="thqm",
        type=str,
        help="Authentication username, only used if a PASSWORD is provided.",
    )
    parser.add_argument(
        "-s", "--seperator", default="\n", help="Entry seperator pattern.", type=str
    )
    parser.add_argument(
        "-o",
        "--oneshot",
        action="store_true",
        default=False,
        help="Shutdown server after first click.",
    )
    parser.add_argument("-t", "--title", default="thqm", help="Page title.", type=str)
    parser.add_argument(
        "--no-shutdown",
        action="store_true",
        default=False,
        help="Remove server shutdown button.",
    )
    parser.add_argument(
        "--no-qrcode",
        action="store_true",
        default=not PYQRCODE_IMPORT,
        help="Remove qrcode button.",
    )
    parser.add_argument(
        "--style",
        default="default",
        choices=[style.name for style in get_styles()],
        type=str,
        help="Page style.",
    )
    args = parser.parse_args()

    # init the configuration folder
    init_conf_folder()

    # get the base dir of the style.
    base_dir = style_base_dir(args.style)

    # if qr code and pyqrcode not installed
    if not PYQRCODE_IMPORT and (args.show_qrcode or not args.no_qrcode):
        print(
            "'pyqrcode' not installed. To install 'pip install pyqrcode'.",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.show_qrcode or not args.no_qrcode:
        qr_path = base_dir / "static/qr_code.svg"
        qr = generate_qr(
            username=args.username,
            password=args.password,
            port=args.port,
            qr_path=qr_path,
        )
        if args.show_qrcode:
            echo(qr.terminal())
    else:
        qr_path = None

    try:
        start_server(
            events=[e.strip() for e in sys.stdin.read().split(args.seperator) if e],
            port=args.port,
            username=args.username,
            password=args.password,
            oneshot=args.oneshot,
            base_dir=base_dir,
            qrcode_button=not args.no_qrcode,
            shutdown_button=not args.no_shutdown,
            title=args.title,
        )
    except KeyboardInterrupt:
        sys.exit(130)
    except BrokenPipeError:
        sys.exit(0)
    except Exception as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    finally:
        # cleanup the qr_code.svg once done
        if qr_path is not None and qr_path.is_file():
            qr_path.unlink()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['thqm']

package_data = \
{'': ['*'],
 'thqm': ['styles/default/static/*',
          'styles/default/static/css/*',
          'styles/default/static/js/*',
          'styles/default/templates/*']}

install_requires = \
['jinja2>=2.11.2,<3.0.0']

extras_require = \
{'qrcode': ['pyqrcode>=1.2.1,<2.0.0']}

entry_points = \
{'console_scripts': ['thqm = thqm.__cli__:main']}

setup_kwargs = {
    'name': 'thqm',
    'version': '1.2.0',
    'description': 'remote command execution made easy.',
    'long_description': '<h1 align="center">thqm</h1>\n<h3 align="center"><img src="https://i.imgur.com/gVB270Z.png" width="150"></h3>\n<h5 align="center">Remote command execution made easy.</h5>\n\n<p align="center">\n  <a href="https://github.com/loiccoyle/thqm/actions?query=workflow%3Atests"><img src="https://github.com/loiccoyle/thqm/workflows/tests/badge.svg"></a>\n  <a href="https://pypi.org/project/thqm/"><img src="https://img.shields.io/pypi/v/thqm"></a>\n  <a href="./LICENSE.md"><img src="https://img.shields.io/badge/license-MIT-blue.svg"></a>\n  <img src="https://img.shields.io/badge/platform-linux%20%7C%20macOS%20%7C%20windows-informational">\n</p>\n<img src="https://i.imgur.com/OrK36nl.png?1" align=\'right\' width=\'170px\'>\n\n> `thqm` takes its name from the arabic تحكم, pronounced tahakum, meaning control.\n\n`thqm` makes it very easy to setup a simple remote control interface on the host machine.\n\n`thqm` is a nifty little HTTP server which reads from standard input. It dynamically generates a simple button menu based on the provided `stdin` and outputs any button the user presses to `stdout`.\nIn a sense its kind of like the [`dmenu`](https://tools.suckless.org/dmenu/)/[`rofi`](https://github.com/davatorium/rofi) of HTTP servers.\n\nThis makes it very flexible and script friendly. See the [examples](./examples) folder for some scripts.\n\n&nbsp;\n\n&nbsp;\n\n# Installation\n```shell\npip install thqm\n```\n\nor with `pyqrcode`\n\n```shell\npip install thqm[qrcode]\n```\n\n`thqm` should work on linux, MacOS and Windows.\n\nIt usually is a good idea to use a virtual environment, or maybe consider using [pipx](https://github.com/pipxproject/pipx).\n\n# Dependencies\n`thqm` requires the following to run:\n  * `python3`\n  * `jinja`\n\nOptional:\n  * `pyqrcode` for qrcode generation.\n\n# Configuration\n`thqm` will create a config folder:\n  * Linux: `$XDG_CONFIG_HOME/thqm` (or `$HOME/.config/thqm` if `$XDG_CONFIG_HOME` is not set)\n  * MacOS: `~/Library/Application Support/thqm`\n  * Windows: `%LOCALAPPDATA%/thqm` (or `~/thqm`)\n\nThis folder holds `thqm`\'s custom styles. A bare bone example, `pure_html`, will be created.\n\nTo add your own custom style, follow the folder structure of the provided example. Maybe have a look at the [`default`](https://github.com/loiccoyle/thqm/tree/master/thqm/styles/default) style.\n\n**Note:** the base folder of the server will the style\'s folder. So to access files in the `static` folder from your `index.html`:\n\n```html\n<link rel="stylesheet" type="text/css" href="static/index.css">\n```\n\n\n# Usage\nCheck the [examples](./examples) folder for some usage examples.\n\n```\n$ thqm --help\n\nRemote command execution made easy.\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -p PORT, --port PORT  Port number. (default: 8901)\n  -q, --show-qrcode     Show the qrcode in terminal, requires "pyqrcode". (default: False)\n  -pw PASSWORD, --password PASSWORD\n                        Authentication password. (default: None)\n  -u USERNAME, --username USERNAME\n                        Authentication username, only used if a PASSWORD is provided. (default:\n                        thqm)\n  -s SEPERATOR, --seperator SEPERATOR\n                        Entry seperator pattern. (default: )\n  -o, --oneshot         Shutdown server after first click. (default: False)\n  -t TITLE, --title TITLE\n                        Page title. (default: thqm)\n  --no-shutdown         Remove server shutdown button. (default: False)\n  --no-qrcode           Remove qrcode button. (default: False)\n  --style {default,pure_html}\n                        Page style. (default: default)\n```\nUse the `-u` and `-pw` arguments to set a username and password to restrict access. The authentication is handled with [HTTP basic authentication](https://en.wikipedia.org/wiki/Basic_access_authentication).\n\nWith the `-s` argument you can define the pattern on which to split `stdin`.\n\nThe `-o` flag will stop the server after the first button press.\n\nThe `-q` (requires `pyqrcode`) flag will print a qr-code in the terminal, this qr-code contains the credentials so it will bypass any authentication, the same is true for the in browser qr-code. This makes it particularly easy to share access with others.\n\nUse `-t` to change the page title.\n\n`--no-shutdown` removes the shutdown server button.\n\n`--no-qrcode` removes the qrcode button.\n\nSelect the page style using the `--style` argument. You can add custom styles in `thqm`\'s config folder.\n',
    'author': 'Loic Coyle',
    'author_email': 'loic.coyle@hotmail.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/loiccoyle/thqm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
